import queue
import threading
import time

import sounddevice as sd
import numpy as np
from PySide6.QtCore import QThread

from airunner.enums import SignalCode
from airunner.settings import SLEEP_TIME_IN_MS
from airunner.workers.worker import Worker


class AudioCaptureWorker(Worker):
    """
    This class is responsible for capturing audio from the microphone.
    It will capture audio when it detects voice activity and then send the audio to the audio_processor_worker.
    """

    def __init__(self):
        super().__init__(signals=(
            (SignalCode.AUDIO_CAPTURE_WORKER_RESPONSE_SIGNAL, self.on_AudioCaptureWorker_response_signal),
            (SignalCode.STT_START_CAPTURE_SIGNAL, self.on_stt_start_capture_signal),
            (SignalCode.STT_STOP_CAPTURE_SIGNAL, self.on_stt_stop_capture_signal),
        ))
        self.listening: bool = False
        self.voice_input_start_time: time.time = None
        self.chunk_duration = self.stt_settings.chunk_duration  # duration of chunks in milliseconds
        self.fs = self.stt_settings.fs
        self.stream = None
        self.running = False
        self._audio_process_queue = queue.Queue()
        self._capture_thread = None

    def on_AudioCaptureWorker_response_signal(self, message: dict):
        item: np.ndarray = message["item"]
        self.logger.debug("Heard signal")
        self.add_to_queue(item)

    def on_stt_start_capture_signal(self):
        if self._capture_thread is not None and self._capture_thread.is_alive():
            return
        self._capture_thread = threading.Thread(target=self._start_listening)
        self._capture_thread.start()

    def on_stt_stop_capture_signal(self):
        if self._capture_thread is not None and self._capture_thread.is_alive():
            self._stop_listening()

    def start(self):
        self.logger.debug("Starting audio capture worker")
        self.running = True
        chunk_duration = self.stt_settings.chunk_duration
        fs = self.stt_settings.fs
        volume_input_threshold = self.stt_settings.volume_input_threshold
        silence_buffer_seconds = self.stt_settings.silence_buffer_seconds
        voice_input_start_time = None
        recording = []
        is_receiving_input = False
        while self.running:
            while self.listening and self.running and self.stream:
                try:
                    chunk, overflowed = self.stream.read(int(chunk_duration * fs))
                except sd.PortAudioError as e:
                    QThread.msleep(SLEEP_TIME_IN_MS)
                    continue
                if np.max(np.abs(chunk)) > volume_input_threshold:  # check if chunk is not silence
                    is_receiving_input = True
                    self.emit_signal(SignalCode.INTERRUPT_PROCESS_SIGNAL)
                    voice_input_start_time = time.time()
                elif voice_input_start_time is not None:
                    # make voice_end_time silence_buffer_seconds after voice_input_start_time
                    end_time = voice_input_start_time + silence_buffer_seconds
                    if time.time() >= end_time:
                        if len(recording) > 0:
                            self.emit_signal(
                                SignalCode.AUDIO_CAPTURE_WORKER_RESPONSE_SIGNAL,
                                {
                                    "item": b''.join(recording)
                                }
                            )
                            recording = []
                            is_receiving_input = False
                if is_receiving_input:
                    chunk_bytes = np.int16(chunk * 32767).tobytes()  # convert to bytes
                    recording.append(chunk_bytes)

            while not self.listening and self.running:
                QThread.msleep(SLEEP_TIME_IN_MS)

    def _start_listening(self):
        self.logger.debug("Start listening")
        self.listening = True
        fs = self.stt_settings.fs
        channels = self.stt_settings.channels
        if self.stream is None:
            self.stream = sd.InputStream(samplerate=fs, channels=channels)

        try:
            self.stream.start()
        except Exception as e:
            self.logger.error(e)

    def _stop_listening(self):
        self.logger.debug("Stop listening")
        self.listening = False
        self.running = False
        try:
            self.stream.stop()
        except Exception as e:
            self.logger.error(e)
        try:
            self.stream.close()
        except Exception as e:
            self.logger.error(e)
        self._capture_thread.join()



from PySide6.QtWidgets import QFileDialog
from PySide6.QtCore import Slot

from airunner.settings import BASE_PATH
from airunner.utils.os.create_airunner_directory import create_airunner_paths
from airunner.windows.setup_wizard.base_wizard import BaseWizard
from airunner.windows.setup_wizard.path_settings.templates.path_settings_ui import Ui_PathSettings


class PathSettings(BaseWizard):
    class_name_ = Ui_PathSettings

    def __init__(self, *args):
        super(PathSettings, self).__init__(*args)

        self.ui.base_path.setText(BASE_PATH)

    @Slot()
    def browse_files(self):
        base_path = str(self.path_settings.base_path)
        file_dialog = QFileDialog()
        file_dialog.setFileMode(QFileDialog.Directory)
        file_dialog.setOption(QFileDialog.ShowDirsOnly)
        file_dialog.setDirectory(base_path)
        file_dialog.setModal(True)
        file_dialog.exec_()
        selected_path = file_dialog.selectedFiles()
        if selected_path:
            self.ui.base_path.setText(selected_path[0])

    def save_settings(self):
        self.update_path_settings("base_path", self.ui.base_path.text())

        """
        Create the airunner paths
        """
        create_airunner_paths(self.path_settings)
        self.update_application_settings("paths_initialized", True)



import base64
import io
import PIL
from PIL import Image


def convert_base64_to_image(base_64_image) -> Image:
    if base_64_image is None:
        return base_64_image
    try:
        decoded = base64.b64decode(base_64_image)
        bytes_ = io.BytesIO(decoded)
        return Image.open(bytes_)
    except PIL.UnidentifiedImageError as e:
        return None



import datetime
import json
import os
import sqlite3
import time
import traceback
from typing import AnyStr
import torch
from PySide6.QtCore import QObject

from llama_index.core import Settings
from llama_index.readers.file import EpubReader, PDFReader, MarkdownReader
from llama_index.core import SimpleDirectoryReader
from llama_index.core.node_parser import SentenceSplitter
from llama_index.core import PromptHelper
from llama_index.core.chat_engine import ContextChatEngine
from llama_index.core import SimpleKeywordTableIndex
from llama_index.core.indices.keyword_table import KeywordTableSimpleRetriever

from airunner.handlers.llm.huggingface_llm import HuggingFaceLLM
from airunner.handlers.llm.custom_embedding import CustomEmbedding
from airunner.handlers.llm.agent.html_file_reader import HtmlFileReader
from airunner.handlers.llm.agent.external_condition_stopping_criteria import ExternalConditionStoppingCriteria
from airunner.handlers.logger import Logger
from airunner.mediator_mixin import MediatorMixin
from airunner.enums import (
    SignalCode,
    LLMChatRole,
    LLMActionType,
    AgentState
)
from airunner.utils.get_torch_device import get_torch_device
from airunner.utils.create_worker import create_worker
from airunner.utils.prepare_llm_generate_kwargs import prepare_llm_generate_kwargs
from airunner.windows.main.settings_mixin import SettingsMixin
from airunner.workers.agent_worker import AgentWorker


class BaseAgent(
    QObject,
    MediatorMixin,
    SettingsMixin
):
    def __init__(self, *args, **kwargs):
        self.logger = Logger(prefix=self.__class__.__name__)
        MediatorMixin.__init__(self)
        SettingsMixin.__init__(self)
        self.model = kwargs.pop("model", None)
        self.__documents = None
        self.__document_reader: SimpleDirectoryReader = None
        self.__index = None
        self.__chat_engine = None
        self.__retriever = None
        self.__storage_context = None
        self.__transformations = None
        self.__index_struct = None
        self.__callback_manager = None
        self.__pdf_reader = None
        self.__epub_reader = None
        self.__html_reader = None
        self.__markdown_reader = None
        self.__embedding = None
        self.__model_name = os.path.expanduser(
            os.path.join(
                self.path_settings.base_path,
                "text/models",
                "sentence_transformers/sentence-t5-large"
            )
        )
        self.__query_instruction = "Search through all available texts and provide a brief summary of the key points which are relevant to the query."
        self.__text_instruction = "Summarize and provide a brief explanation of the text. Stay concise and to the point."
        self.__state = AgentState.SEARCH
        self.__chunk_size = 1000
        self.__chunk_overlap = 512
        self.__target_files = []
        self.__rag_model = None
        self.__rag_tokenizer = None
        self.__model = None
        self.__tokenizer = None
        self.__llm = None
        self._chatbot = None
        self.action = LLMActionType.CHAT
        self.rendered_template = None
        self.tokenizer = kwargs.pop("tokenizer", None)
        self.streamer = kwargs.pop("streamer", None)
        self.chat_template = kwargs.pop("chat_template", "")
        self.is_mistral = kwargs.pop("is_mistral", True)
        self.conversation_id = None
        self.history = self.db_handler.load_history_from_db(self.conversation_id)  # Load history by conversation ID
        super().__init__(*args, **kwargs)
        self.prompt = ""
        self.thread = None
        self.do_interrupt = False
        self.response_worker = create_worker(AgentWorker)
        self.load_rag(model=self.model, tokenizer=self.tokenizer)

    @property
    def available_actions(self):
        return {
            0: LLMActionType.QUIT_APPLICATION,
            1: LLMActionType.TOGGLE_FULLSCREEN,
            2: LLMActionType.TOGGLE_TTS,
            3: LLMActionType.GENERATE_IMAGE,
            4: LLMActionType.PERFORM_RAG_SEARCH,
            5: LLMActionType.CHAT,
        }

    @property
    def username(self) -> str:
        return self.chatbot.username

    @property
    def botname(self) -> str:
        return self.chatbot.botname

    @property
    def bot_mood(self) -> str:
        return self.chatbot.bot_mood

    @bot_mood.setter
    def bot_mood(self, value: str):
        chatbot = self.chatbot
        chatbot.bot_mood = value
        self.db_handler.save_object(chatbot)
        self.emit_signal(SignalCode.BOT_MOOD_UPDATED)

    @property
    def bot_personality(self) -> str:
        return self.chatbot.bot_personality

    def unload(self):
        self.unload_rag()
        del self.model
        del self.tokenizer
        self.model = None
        self.tokenizer = None
        self.thread = None

    def clear_history(self):
        self.history = []
        self.reload_rag()
        self.conversation_id = None

    def update_conversation_title(self, title):
        self.db_handler.update_conversation_title(self.conversation_id, title)

    def create_conversation(self):
        # Get the most recent conversation ID
        recent_conversation_id = self.db_handler.get_most_recent_conversation_id()

        # Check if there are messages for the most recent conversation ID
        if recent_conversation_id is not None:
            messages = self.db_handler.load_history_from_db(recent_conversation_id)
            if not messages:
                self.conversation_id = recent_conversation_id
                return

        # If there are messages or no recent conversation ID, create a new conversation
        self.conversation_id = self.db_handler.create_conversation()

    def interrupt_process(self):
        self.do_interrupt = True

    def do_interrupt_process(self):
        interrupt = self.do_interrupt
        self.do_interrupt = False
        return interrupt

    @property
    def use_cuda(self):
        return torch.cuda.is_available()

    @property
    def cuda_index(self):
        return 0

    def mood(self, botname: str, bot_mood: str, use_mood: bool) -> str:
        return (
            f"{botname} is not a computer or a bot, you are a human with emotions. "
            f"{botname}'s mood changes depending on the conversation and other factors.\n"
            f"{botname}'s current mood: {bot_mood}."
            #f"{botname}'s evaluation of {username}: {self.user_evaluation}"
        ) if (use_mood and bot_mood != "") else ""

    def history_prompt(self) -> str:
        history = ""
        previous_date = None
        for message in self.history:
            # Check if the timestamp is a string and convert it to a datetime object if necessary
            if isinstance(message["timestamp"], str):
                current_date = datetime.datetime.strptime(message["timestamp"], "%Y-%m-%d %H:%M:%S").date()
            else:
                current_date = message["timestamp"].date()

            if previous_date is None or current_date != previous_date:
                history += f"\n--- {current_date.strftime('%B %d, %Y')} ---\n"
                previous_date = current_date
            name = ""
            if message["role"] == LLMChatRole.HUMAN.value:
                name = self.username
            elif message["role"] == LLMChatRole.ASSISTANT.value:
                name = self.botname
            history += f"{name}: {message['content']}\n"
        return (
            "------\n"
            "Chat History:\n"
            f"{history}"
        )

    def date_time_prompt(self) -> str:
        if not self.chatbot.use_datetime:
            return ""
        current_date = datetime.datetime.now().strftime("%A %b %d, %Y")
        current_time = datetime.datetime.now().strftime("%I:%M:%S %p")
        current_timezone = time.tzname
        prompt = [
            "\n======\n",
            f"Use the following information to help you with your response, but do not include it in your response or reference it directly unless asked.",
            f"Date: {current_date}, Time: {current_time}, Timezone: {current_timezone}",
        ]
        return "\n".join(prompt)

    def build_system_prompt(
        self,
        action
    ):
        system_instructions = ""
        guardrails_prompt = ""
        use_mood = self.chatbot.use_mood
        use_personality = self.chatbot.use_personality
        use_names = self.chatbot.assign_names
        use_system_instructions = self.chatbot.use_system_instructions
        use_guardrails = self.chatbot.use_guardrails
        bot_mood = self.bot_mood
        bot_personality = self.chatbot.bot_personality
        username = self.chatbot.username
        botname = self.chatbot.botname
        if use_system_instructions:
            system_instructions = self.chatbot.system_instructions
        if use_guardrails:
            guardrails_prompt = self.chatbot.guardrails_prompt

        system_prompt = []

        if action is LLMActionType.CHAT:
            """
            Build the system prompt for chat template
            """
            system_prompt = [
                guardrails_prompt,
                system_instructions,
                self.names_prompt(use_names, botname, username),
                self.mood(botname, bot_mood, use_mood),
                self.personality_prompt(bot_personality, use_personality),
                self.history_prompt(),
                self.date_time_prompt()
            ]

        elif action is LLMActionType.GENERATE_IMAGE:
            prompt_template = self.get_prompt_template_by_name("image")
            # system_prompt = [
            #     prompt_template.guardrails,
            #     prompt_template.system,
            #     self.history_prompt()
            # ]
            system_prompt = [
                (
                    "You are an image generator. "
                    "You will be provided with a JSON string and it is your goal to replace the PLACEHOLDER "
                    "text with text appropriate for the given attribute in the JSON string. "
                    "You will follow all of the rules to generate descriptions for an image. "
                    "\n------\n"
                    "RULES:\n"
                    "When available, use the Additional Context to keep your generated content in line with the existing context.\n"
                    "You will be given instructions on what type of image to generate and you will do your best to follow those instructions.\n"
                    "You will only generate a value for the given attribute.\n"
                    "Never respond in a conversational manner. Never provide additional information, details or information.\n"
                    "You will only provide the requested information by replacing the PLACEHOLDER.\n"
                    "Never change the attribute\n"
                    "You must not change the structure of the data.\n"
                    "You will only return JSON strings.\n"
                    "You will not return any other data types.\n"
                    "You are an artist, so use your imagination and keep things interesting.\n"
                    "You will not respond in a conversational manner or with additonal notes or information.\n"
                    f"Only return one JSON block. Do not generate instructions or additional information.\n"
                    "You must never break the rules.\n"
                    "Here is a description of the attributes: \n"
                    "`description`: This should describe the overall subject and look and feel of the image\n"
                    "`composition`: This should describe the attributes of the image such as color, composition and other details\n"
                ),
                self.history_prompt()
            ]

        elif action is LLMActionType.APPLICATION_COMMAND:
            prompt_template = self.get_prompt_template_by_name("application_command")
            system_instructions = prompt_template.system

            # Create a list of commands that the bot can choose from
            for index, action in self.available_actions.items():
                system_instructions += f"{index}: {action.value}\n"

            system_prompt = [
                self.names_prompt(use_names, botname, username),
                self.mood(botname, bot_mood, use_mood),
                system_instructions,
                "------\n",
                "Chat History:\n",
                f"{self.username}: {self.prompt}\n",
            ]

        elif action is LLMActionType.SUMMARIZE:
            prompt_template = self.get_prompt_template_by_name("summarize")
            system_instructions = prompt_template.system
            system_prompt = [
                system_instructions,
                self.history_prompt()
            ]

        elif action is LLMActionType.UPDATE_MOOD:
            prompt_template = self.get_prompt_template_by_name("update_mood")
            system_instructions = prompt_template.system
            system_prompt = [
                guardrails_prompt,
                system_instructions,
                self.names_prompt(use_names, botname, username),
                self.mood(botname, bot_mood, use_mood),
                self.personality_prompt(bot_personality, use_personality),
                self.history_prompt(),
            ]

        elif action is LLMActionType.PERFORM_RAG_SEARCH:
            prompt_template = self.get_prompt_template_by_name("rag_search")
            system_instructions = prompt_template.system
            system_prompt = [
                guardrails_prompt,
                system_instructions,
                self.names_prompt(use_names, botname, username),
                self.mood(botname, bot_mood, use_mood),
                self.personality_prompt(bot_personality, use_personality),
                self.history_prompt(),
            ]

        elif action is LLMActionType.QUIT_APPLICATION:
            self.emit_signal(SignalCode.QUIT_APPLICATION)

        elif action is LLMActionType.TOGGLE_FULLSCREEN:
            self.emit_signal(SignalCode.TOGGLE_FULLSCREEN_SIGNAL)

        return "\n".join(system_prompt)

    def names_prompt(self, use_names: bool, botname: str, username: str) -> str:
        return f"Your name is {botname}. \nThe user's name is {username}." if use_names else ""

    def personality_prompt(self, bot_personality: str, use_personality: bool) -> str:
        return (
            f"Your personality: {bot_personality}."
        ) if use_personality else ""

    def prepare_messages(
        self,
        action
    ) -> list:
        system_prompt = self.build_system_prompt(action)
        if action is LLMActionType.APPLICATION_COMMAND:
            prompt = (
                "Choose an action from THE LIST of commands for the text above. "
                "Only return the number of the command."
            )
        elif action is LLMActionType.GENERATE_IMAGE:
            prompt = (
                f"Replace the placeholder values in the following JSON:\n"
                "```json\n"+ json.dumps(dict(
                    description="PLACEHOLDER",
                    composition="PLACEHOLDER"
                )) +"\n```\n"
            )
        elif action is LLMActionType.SUMMARIZE:
            prompt = (
                f"Summarize the conversation history"
            )
        else:
            prompt = f"Respond to {self.username}"
        messages = [
            {
                "content": system_prompt,
                "role": LLMChatRole.SYSTEM.value
            },
            {
                "content": prompt,
                "role": LLMChatRole.HUMAN.value
            }
        ]
        return messages

    def get_rendered_template(
        self,
        action: LLMActionType
    ) -> str:
        conversation = self.prepare_messages(action)
        rendered_template = self.tokenizer.apply_chat_template(
            chat_template=self.chat_template,
            conversation=conversation,
            tokenize=False
        )

        # HACK: current version of transformers does not allow us to pass
        # variables to the chat template function, so we apply those here
        variables = {
            "speaker_name": self.botname,
            "listener_name": self.username,
            "username": self.username,
            "botname": self.botname,
            "bot_mood": self.bot_mood,
            "bot_personality": self.bot_personality,
        }
        for key, value in variables.items():
            value = value or ""
            rendered_template = rendered_template.replace("{{ " + key + " }}", value)
        return rendered_template

    @property
    def override_parameters(self):
        generate_kwargs = prepare_llm_generate_kwargs(self.llm_generator_settings)
        return generate_kwargs if self.llm_generator_settings.override_parameters else {}

    @property
    def system_instructions(self):
        return self.chatbot.system_instructions

    @property
    def generator_settings(self) -> dict:
        return prepare_llm_generate_kwargs(self.chatbot)

    @property
    def device(self):
        return get_torch_device(self.memory_settings.default_gpu_llm)

    @property
    def target_files(self):
        return [
            target_file.file_path for target_file in self.chatbot.target_files
        ]

    @property
    def query_instruction(self):
        if self.__state == AgentState.SEARCH:
            return self.__query_instruction
        elif self.__state == AgentState.CHAT:
            return "Search through the chat history for anything relevant to the query."

    @property
    def text_instruction(self):
        if self.__state == AgentState.SEARCH:
            return self.__text_instruction
        elif self.__state == AgentState.CHAT:
            return "Use the text to respond to the user"

    @property
    def index(self):
        if self.__state == AgentState.SEARCH:
            return self.__index
        elif self.__state == AgentState.CHAT:
            return self.__chat_history_index

    @property
    def llm(self):
        if self.__llm is None:
            try:
                if self.llm_generator_settings.use_api:
                    self.__llm = self.__model
                else:
                    self.__llm = HuggingFaceLLM(model=self.__model, tokenizer=self.__tokenizer)
            except Exception as e:
                self.logger.error(f"Error loading LLM: {str(e)}")
        return self.__llm

    @property
    def chat_engine(self):
        return self.__chat_engine

    @property
    def is_llama_instruct(self):
        return True

    def run(
        self,
        prompt: str,
        action: LLMActionType,
        **kwargs
    ):
        self.action = action

        if action is LLMActionType.TOGGLE_TTS:
            self.emit_signal(SignalCode.TOGGLE_TTS_SIGNAL)
            return

        self.logger.debug("Running...")
        self.prompt = prompt
        streamer = self.streamer

        if self.conversation_id is None:
            self.create_conversation()

        # Add the user's message to history
        if action in (
            LLMActionType.CHAT,
            LLMActionType.PERFORM_RAG_SEARCH,
            LLMActionType.GENERATE_IMAGE,
        ):
            self.add_message_to_history(self.prompt, LLMChatRole.HUMAN)

        self.rendered_template = self.get_rendered_template(action)

        model_inputs = self.tokenizer(
            self.rendered_template,
            return_tensors="pt"
        ).to(self.device)

        kwargs.update(
            streamer=streamer,
            action=action,
            do_emit_response=True
        )

        if streamer:
            self.run_with_thread(
                model_inputs,
                **kwargs
            )
        else:
            self.emit_signal(SignalCode.UNBLOCK_TTS_GENERATOR_SIGNAL)
            stopping_criteria = ExternalConditionStoppingCriteria(self.do_interrupt_process)
            data = self.prepare_generate_data(model_inputs, stopping_criteria)
            res = self.model.generate(**data)
            response = self.tokenizer.decode(res[0])
            self.emit_signal(
                SignalCode.LLM_TEXT_STREAMED_SIGNAL,
                dict(
                    message=response,
                    is_first_message=True,
                    is_end_of_message=True,
                    name=self.botname,
                )
            )

            return response

    def prepare_generate_data(self, model_inputs, stopping_criteria):
        data = dict(
            **model_inputs,
            **self.generator_settings,
            stopping_criteria=[stopping_criteria]
        )
        data.update(self.override_parameters)
        return data

    def run_with_thread(
        self,
        model_inputs,
        action: LLMActionType,
        **kwargs,
    ):
        # Generate the response
        self.logger.debug("Generating...")

        self.emit_signal(SignalCode.UNBLOCK_TTS_GENERATOR_SIGNAL)

        stopping_criteria = ExternalConditionStoppingCriteria(self.do_interrupt_process)

        data = self.prepare_generate_data(model_inputs, stopping_criteria)
        streamer = kwargs.get("streamer", self.streamer)
        data["streamer"] = streamer

        # if "attention_mask" in data:
        #     del data["attention_mask"]

        self.do_interrupt = False

        if action is not LLMActionType.PERFORM_RAG_SEARCH:
            try:
                self.response_worker.add_to_queue({
                    "model": self.model,
                    "kwargs": data,
                    "prompt": self.prompt,
                    "botname": self.botname,
                })
            except Exception as e:
                self.logger.error("545: An error occurred in model.generate:")
                self.logger.error(str(e))
                self.logger.error(traceback.format_exc())

        # strip all new lines from rendered_template:
        #self.rendered_template = self.rendered_template.replace("\n", " ")
        eos_token = self.tokenizer.eos_token
        bos_token = self.tokenizer.bos_token
        if self.is_mistral:
            self.rendered_template = bos_token + self.rendered_template
        skip = True
        streamed_template = ""
        replaced = False
        is_end_of_message = False
        is_first_message = True

        if action == LLMActionType.GENERATE_IMAGE:
            self.emit_signal(
                SignalCode.LLM_TEXT_STREAMED_SIGNAL,
                dict(
                    message="Generating image prompt.\n",
                    is_first_message=is_first_message,
                    is_end_of_message=False,
                    name=self.botname,
                )
            )
            is_first_message = False

        if streamer and action in (
            LLMActionType.CHAT,
            LLMActionType.GENERATE_IMAGE,
            LLMActionType.UPDATE_MOOD,
            LLMActionType.SUMMARIZE,
            LLMActionType.APPLICATION_COMMAND
        ):
            for new_text in streamer:
                # strip all newlines from new_text
                streamed_template += new_text
                if self.is_mistral:
                    streamed_template = streamed_template.replace(f"{bos_token} [INST]", f"{bos_token}[INST]")
                    streamed_template = streamed_template.replace("  [INST]", " [INST]")
                # iterate over every character in rendered_template and
                # check if we have the same character in streamed_template
                if not replaced:
                    for i, char in enumerate(self.rendered_template):
                        try:
                            if char == streamed_template[i]:
                                skip = False
                            else:
                                skip = True
                                break
                        except IndexError:
                            skip = True
                            break
                if skip:
                    continue
                elif not replaced:
                    replaced = True
                    streamed_template = streamed_template.replace(self.rendered_template, "")
                else:
                    if eos_token in new_text:
                        streamed_template = streamed_template.replace(eos_token, "")
                        new_text = new_text.replace(eos_token, "")
                        is_end_of_message = True
                    # strip botname from new_text
                    new_text = new_text.replace(f"{self.botname}:", "")
                    if action in (
                        LLMActionType.CHAT,
                        LLMActionType.PERFORM_RAG_SEARCH,
                        LLMActionType.GENERATE_IMAGE
                    ):
                        self.emit_signal(
                            SignalCode.LLM_TEXT_STREAMED_SIGNAL,
                            dict(
                                message=new_text,
                                is_first_message=is_first_message,
                                is_end_of_message=is_end_of_message,
                                name=self.botname,
                            )
                        )
                    else:
                        self.emit_signal(
                            SignalCode.LLM_TEXT_STREAMED_SIGNAL,
                            dict(
                                message="",
                                is_first_message=is_first_message,
                                is_end_of_message=is_end_of_message,
                                name=self.botname,
                            )
                        )
                    is_first_message = False
        elif action is LLMActionType.PERFORM_RAG_SEARCH:
            streamed_template = ""
            data = dict(
                **self.generator_settings,
                stopping_criteria=[stopping_criteria]
            )
            data.update(self.override_parameters)
            self.llm.generate_kwargs = data
            response = self.chat_engine.stream_chat(message=self.prompt)
            is_first_message = True
            is_end_of_message = False
            for new_text in response.response_gen:
                streamed_template += new_text

                self.emit_signal(
                    SignalCode.LLM_TEXT_STREAMED_SIGNAL,
                    dict(
                        message=" " +new_text,
                        is_first_message=is_first_message,
                        is_end_of_message=is_end_of_message,
                        name=self.botname,
                    )
                )
                is_first_message = False

        if streamed_template is not None:
            if action is LLMActionType.CHAT:
                self.add_message_to_history(
                    streamed_template,
                    LLMChatRole.ASSISTANT
                )

            elif action is LLMActionType.UPDATE_MOOD:
                self.bot_mood = streamed_template
                return self.run(
                    prompt=self.prompt,
                    action=LLMActionType.CHAT,
                    **kwargs,
                )

            elif action is LLMActionType.SUMMARIZE:
                self.update_conversation_title(streamed_template)

            elif action is LLMActionType.GENERATE_IMAGE:
                self.emit_signal(
                    SignalCode.LLM_IMAGE_PROMPT_GENERATED_SIGNAL,
                    {
                        "message": streamed_template,
                        "type": "photo"
                    }
                )
            elif action is LLMActionType.APPLICATION_COMMAND:
                index = ''.join(c for c in streamed_template if c.isdigit())
                try:
                    index = int(index)
                except ValueError:
                    index = 0

                try:
                    action = self.available_actions[index]
                except KeyError:
                    action = LLMActionType.CHAT

                if action is not None:
                    return self.run(
                        prompt=self.prompt,
                        action=action,
                    )
        return streamed_template

    def add_chatbot_response_to_history(self, response: dict):
        self.add_message_to_history(
            response["message"],
            response["role"]
        )

    def get_db_connection(self):
        return sqlite3.connect('airunner.db')

    def add_message_to_history(
        self,
        content: AnyStr,
        role: LLMChatRole = LLMChatRole.ASSISTANT
    ):
        if content is None:
            return

        name = self.username
        is_bot = False
        if role is LLMChatRole.ASSISTANT and content:
            content = content.replace(f"{self.botname}:", "")
            content = content.replace(f"{self.botname}", "")
            is_bot = True
            name = self.botname

        self.history.append({
            "role": role.value,
            "content": content,
            "name": name,
            "is_bot": is_bot,
            "timestamp": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            "conversation_id": self.conversation_id  # Use the stored conversation ID
        })

        self.db_handler.add_message_to_history(
            content,
            role.value,
            name,
            is_bot,
            self.conversation_id
        )

    def on_load_conversation(self, message):
        self.history = []
        self.conversation_id = message["conversation_id"]
        self.history = self.db_handler.load_history_from_db(self.conversation_id)
        self.emit_signal(SignalCode.SET_CONVERSATION, {
            "messages": self.history
        })

    def load_rag(self, model, tokenizer):
        self.__model = model
        self.__tokenizer = tokenizer
        self.__load_rag()

    def unload_rag(self):
        if self.__llm is not None:
            self.__llm.unload()
            del self.__llm
            self.__llm = None

    def reload_rag(self, data: dict = None):
        self.logger.debug("Reloading RAG index...")
        self.__target_files = data["target_files"] if data is not None else self.__target_files
        self.__load_rag()

    def __load_rag(self):
        self.__load_rag_tokenizer()
        self.__load_rag_model()
        self.__load_embeddings()
        self.__load_readers()
        self.__load_file_extractor()
        self.__load_document_reader()
        self.__load_documents()
        self.__load_text_splitter()
        self.__load_prompt_helper()
        self.__load_settings()
        self.__load_document_index()
        self.__load_retriever()
        self.__load_context_chat_engine()

    def __load_rag_tokenizer(self):
        self.logger.debug("Loading RAG tokenizer...")
        # TODO
        pass

    def __load_rag_model(self):
        self.logger.debug("Loading RAG model...")
        # TODO
        pass

    def __load_embeddings(self):
        self.logger.debug("Loading embeddings...")
        self.__embedding = CustomEmbedding(self.llm)

    def __load_readers(self):
        self.__pdf_reader = PDFReader()
        self.__epub_reader = EpubReader()
        self.__html_reader = HtmlFileReader()
        self.__markdown_reader = MarkdownReader()

    def __load_file_extractor(self):
        self.file_extractor = {
            ".pdf": self.__pdf_reader,
            ".epub": self.__epub_reader,
            ".html": self.__html_reader,
            ".htm": self.__html_reader,
            ".md": self.__markdown_reader,
        }

    def __load_document_reader(self):
        if self.target_files is None or len(self.target_files) == 0:
            return
        self.logger.debug("Loading document reader...")
        try:
            self.__document_reader = SimpleDirectoryReader(
                input_files=self.target_files,
                file_extractor=self.file_extractor,
                exclude_hidden=False
            )
            self.logger.debug("Document reader loaded successfully.")
        except ValueError as e:
            self.logger.error(f"Error loading document reader: {str(e)}")

    def __load_documents(self):
        if not self.__document_reader:
            return
        self.logger.debug("Loading documents...")
        self.__documents = self.__document_reader.load_data()

    def __load_text_splitter(self):
        self.__text_splitter = SentenceSplitter(
            chunk_size=256,
            chunk_overlap=20
        )

    def __load_prompt_helper(self):
        self.__prompt_helper = PromptHelper(
            context_window=4096,
            num_output=1024,
            chunk_overlap_ratio=0.1,
            chunk_size_limit=None,
        )

    def __load_context_chat_engine(self):
        try:
            self.__chat_engine = ContextChatEngine.from_defaults(
                retriever=self.__retriever,
                chat_history=self.history,
                memory=None,
                system_prompt="Search the full text and find all relevant information related to the query.",
                node_postprocessors=[],
                llm=self.__llm,
            )
        except Exception as e:
            self.logger.error(f"Error loading chat engine: {str(e)}")

    def __load_settings(self):
        Settings.llm = self.__llm
        Settings._embed_model = self.__embedding
        Settings.node_parser = self.__text_splitter
        Settings.num_output = 512
        Settings.context_window = 3900

    def __load_document_index(self):
        self.logger.debug("Loading index...")
        documents = self.__documents or []
        try:
            self.__index = SimpleKeywordTableIndex.from_documents(
                documents,
                llm=self.__llm
            )
            self.logger.debug("Index loaded successfully.")
        except TypeError as e:
            self.logger.error(f"Error loading index: {str(e)}")

    def __load_retriever(self):
        try:
            self.__retriever = KeywordTableSimpleRetriever(
                index=self.__index,
            )
            self.logger.debug("Retriever loaded successfully with index.")
        except Exception as e:
            self.logger.error(f"Error setting up the retriever: {str(e)}")



import os
import threading

import numpy as np
import torch
from transformers.models.whisper.modeling_whisper import WhisperForConditionalGeneration
from transformers.models.whisper.processing_whisper import WhisperProcessor
from transformers.models.whisper.feature_extraction_whisper import WhisperFeatureExtractor

from airunner.handlers.base_handler import BaseHandler
from airunner.enums import SignalCode, ModelType, ModelStatus, LLMChatRole
from airunner.exceptions import NaNException
from airunner.settings import DEFAULT_STT_HF_PATH
from airunner.utils.clear_memory import clear_memory


class WhisperHandler(BaseHandler):
    """
    Handler for the Whisper model from OpenAI.
    """
    def __init__(self, *args, **kwargs):
        self.model_type = ModelType.STT
        self.model_class = "stt"
        super().__init__(*args, **kwargs)
        self._lock = threading.Lock()
        self._model = None
        self._processor = None
        self._feature_extractor = None
        self._fs = 16000

    @property
    def stt_is_loading(self):
        return self.model_status is ModelStatus.LOADING

    @property
    def stt_is_loaded(self):
        return self.model_status is ModelStatus.LOADED

    @property
    def stt_is_unloaded(self):
        return self.model_status is ModelStatus.UNLOADED

    @property
    def model_path(self) -> str:
        file_path = os.path.expanduser(os.path.join(
            self.path_settings.base_path,
            "text",
            "models",
            "stt",
            "openai",
            "whisper-tiny"
        ))
        return os.path.abspath(file_path)

    def process_audio(self, audio_data):
        with self._lock:
            item = audio_data["item"]
            # Convert the byte string to a float32 array
            inputs = np.frombuffer(item, dtype=np.int16)
            inputs = inputs.astype(np.float32) / 32767.0
            try:
                transcription = self._process_inputs(inputs)
            except Exception as e:
                self.logger.error(f"Failed to process inputs {e}")
                self.logger.error(e)
            try:
                self._process_human_speech(transcription)
            except ValueError as e:
                self.logger.error(f"Failed to process audio {e}")

    def load(self):
        if self.stt_is_loading or self.stt_is_loaded:
            return
        self.logger.debug("Loading Whisper (text-to-speech)")
        self.unload()
        self.change_model_status(ModelType.STT, ModelStatus.LOADING)
        self._load_model()
        self._load_processor()
        self._load_feature_extractor()
        if (
            self._model is not None and
            self._processor is not None and
            self._feature_extractor is not None
        ):
            self.change_model_status(ModelType.STT, ModelStatus.LOADED)
            return True
        else:
            self.change_model_status(ModelType.STT, ModelStatus.FAILED)
            return False

    def unload(self):
        if self.stt_is_loading or self.stt_is_unloaded:
            return
        self.logger.debug("Unloading Whisper (text-to-speech)")
        self.change_model_status(ModelType.STT, ModelStatus.LOADING)
        self._unload_model()
        self._unload_processor()
        self._unload_feature_extractor()
        self.change_model_status(ModelType.STT, ModelStatus.UNLOADED)

    def _load_model(self):
        model_path = self.model_path
        self.logger.debug(f"Loading model from {model_path}")
        try:
            self._model = WhisperForConditionalGeneration.from_pretrained(
                model_path,
                local_files_only=True,
                torch_dtype=torch.bfloat16,
                device_map=self.device,
                use_safetensors=True
            )
        except Exception as e:
            self.logger.error(f"Failed to load model")
            self.logger.error(e)
            return None

    def _load_processor(self):
        model_path = self.model_path
        self.logger.debug(f"Loading processor from {model_path}")
        try:
            self._processor = WhisperProcessor.from_pretrained(
                model_path,
                local_files_only=True,
                torch_dtype=torch.bfloat16,
                device_map=self.device
            )
        except Exception as e:
            self.logger.error(f"Failed to load processor")
            self.logger.error(e)
            return None

    def _load_feature_extractor(self):
        model_path = self.model_path
        self.logger.debug(f"Loading feature extractor {model_path}")
        try:
            self._feature_extractor = WhisperFeatureExtractor.from_pretrained(
                model_path,
                local_files_only=True,
                torch_dtype=torch.bfloat16,
                device_map=self.device
            )
        except Exception as e:
            self.logger.error(f"Failed to load feature extractor")
            self.logger.error(e)
            return None

    def _unload_model(self):
        del self._model
        self._model = None
        clear_memory(self.device)

    def _unload_processor(self):
        del self._processor
        self._processor = None
        clear_memory(self.device)

    def _unload_feature_extractor(self):
        del self._feature_extractor
        self._feature_extractor = None
        clear_memory(self.device)

    def _process_inputs(
        self,
        inputs: np.ndarray,
        role: LLMChatRole = LLMChatRole.HUMAN,
    ) -> str:
        inputs = torch.from_numpy(inputs)

        if torch.isnan(inputs).any():
            raise NaNException

        inputs = self._feature_extractor(inputs, sampling_rate=self._fs, return_tensors="pt")
        if torch.isnan(inputs.input_features).any():
            raise NaNException

        inputs["input_features"] = inputs["input_features"].to(torch.bfloat16)
        if torch.isnan(inputs.input_features).any():
            raise NaNException

        inputs = inputs.to(self._model.device)
        if torch.isnan(inputs.input_features).any():
            raise NaNException

        transcription = self._run(inputs, role)
        if transcription is None or 'nan' in transcription:
            raise NaNException

        return transcription

    def _process_human_speech(self, transcription: str = None):
        """
        Process the human speech.
        This method is called when the model has processed the human speech
        and the transcription is ready to be added to the chat history.
        This should only be used for human speech.
        :param transcription:
        :return:
        """
        if transcription == "":
            raise ValueError("Transcription is empty")
        self.logger.debug("Processing human speech")
        data = {
            "message": transcription,
            "role": LLMChatRole.HUMAN
        }
        self.emit_signal(
            SignalCode.ADD_CHATBOT_MESSAGE_SIGNAL,
            data
        )

    def _run(
        self,
        inputs,
        role: LLMChatRole = LLMChatRole.HUMAN,
    ) -> str:
        """
        Run the model on the given inputs.
        :param inputs: str - The transcription of the audio data.
        :param role: LLMChatRole - The role of the speaker.
        :return:
        """
        # Extract the tensor from the BatchFeature object
        self.logger.debug("Running model")
        try:
            input_tensor = inputs.input_values
        except AttributeError:
            input_tensor = inputs.input_features

        if torch.isnan(input_tensor).any():
            raise NaNException

        input_features = inputs.input_features
        if torch.isnan(input_features).any():
            raise NaNException

        generated_ids = self._model.generate(
            input_features=input_features,
            # generation_config=None,
            # logits_processor=None,
            # stopping_criteria=None,
            # prefix_allowed_tokens_fn=None,
            # synced_gpus=True,
            # return_timestamps=None,
            # task="transcribe",
            # language="en",
            # is_multilingual=True,
            # prompt_ids=None,
            # prompt_condition_type=None,
            # condition_on_prev_tokens=None,
            temperature=0.8,
            # compression_ratio_threshold=None,
            # logprob_threshold=None,
            # no_speech_threshold=None,
            # num_segment_frames=None,
            # attention_mask=None,
            # time_precision=0.02,
            # return_token_timestamps=None,
            # return_segments=False,
            # return_dict_in_generate=None,
        )
        if torch.isnan(generated_ids).any():
            raise NaNException

        transcription = self.process_transcription(generated_ids)
        if len(transcription) == 0 or len(transcription.split(" ")) == 1:
            return ""

        # Emit the transcription so that other handlers can use it
        self.emit_signal(SignalCode.AUDIO_PROCESSOR_RESPONSE_SIGNAL, {
            "transcription": transcription,
            "role": role
        })

        return transcription

    def process_transcription(self, generated_ids) -> str:
        # Decode the generated ids
        transcription = self._processor.batch_decode(
            generated_ids,
            skip_special_tokens=True
        )[0]

        # Remove leading and trailing whitespace
        transcription = transcription.strip()

        # Remove any extra whitespace
        transcription = " ".join(transcription.split())

        return transcription



import base64
import io
from PIL import Image
from airunner.handlers.logger import Logger


def convert_image_to_base64(image: Image) -> str:
    if image is None:
        raise ValueError("Image is None")
    img_byte_arr = io.BytesIO()
    try:
        image.save(img_byte_arr, format='PNG')
    except AttributeError as e:
        logger = Logger(prefix="convert_image_to_base64")
        logger.error(f"Something went wrong with image conversion to base64: {e}")
        return None
    img_byte_arr = img_byte_arr.getvalue()
    image_base64 = base64.encodebytes(img_byte_arr).decode('ascii')
    return image_base64



from PySide6.QtCore import QObject, Signal, QThread


class DirectoryWatcher(QObject):
    scan_completed = Signal(bool)

    def __init__(self, base_path: str, scan_function: callable):
        super().__init__()
        self.base_path = base_path
        self._scan_function = scan_function

    def run(self):
        while True:
            force_reload = self._scan_function(self.base_path)
            self.scan_completed.emit(force_reload)
            QThread.sleep(1)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'message.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QFrame, QGridLayout, QHBoxLayout,
    QLabel, QSizePolicy, QTextEdit, QWidget)

class Ui_message(object):
    def setupUi(self, message):
        if not message.objectName():
            message.setObjectName(u"message")
        message.resize(418, 902)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(message.sizePolicy().hasHeightForWidth())
        message.setSizePolicy(sizePolicy)
        message.setMinimumSize(QSize(0, 40))
        self.gridLayout = QGridLayout(message)
        self.gridLayout.setObjectName(u"gridLayout")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.user_name = QLabel(message)
        self.user_name.setObjectName(u"user_name")
        self.user_name.setAlignment(Qt.AlignBottom|Qt.AlignLeading|Qt.AlignLeft)

        self.horizontalLayout.addWidget(self.user_name)

        self.content = QTextEdit(message)
        self.content.setObjectName(u"content")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.MinimumExpanding)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.content.sizePolicy().hasHeightForWidth())
        self.content.setSizePolicy(sizePolicy1)
        self.content.setMinimumSize(QSize(0, 40))
        self.content.setStyleSheet(u"border-radius: 5px; border: 5px solid #1f1f1f; background-color: #1f1f1f; color: #ffffff;")
        self.content.setFrameShape(QFrame.NoFrame)
        self.content.setFrameShadow(QFrame.Plain)
        self.content.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.content.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        self.horizontalLayout.addWidget(self.content)

        self.bot_name = QLabel(message)
        self.bot_name.setObjectName(u"bot_name")
        self.bot_name.setAlignment(Qt.AlignBottom|Qt.AlignLeading|Qt.AlignLeft)

        self.horizontalLayout.addWidget(self.bot_name)


        self.gridLayout.addLayout(self.horizontalLayout, 0, 0, 1, 1)


        self.retranslateUi(message)

        QMetaObject.connectSlotsByName(message)
    # setupUi

    def retranslateUi(self, message):
        message.setWindowTitle(QCoreApplication.translate("message", u"Form", None))
        self.user_name.setText(QCoreApplication.translate("message", u"TextLabel", None))
        self.bot_name.setText(QCoreApplication.translate("message", u"TextLabel", None))
    # retranslateUi




from airunner.enums import SignalCode
from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.llm.templates.message_ui import Ui_message

from PySide6.QtGui import QTextCursor, QFontDatabase, QFont
from PySide6.QtWidgets import QTextEdit
from PySide6.QtGui import QFontMetrics
from PySide6.QtCore import Qt, QSize
from PySide6.QtCore import Signal


class AutoResizingTextEdit(QTextEdit):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.document().contentsChanged.connect(self.sizeChange)


class MessageWidget(BaseWidget):
    widget_class_ = Ui_message
    textChanged = Signal()

    def __init__(self, *args, **kwargs):
        self.name = kwargs.pop("name")
        self.message = kwargs.pop("message")
        self.is_bot = kwargs.pop("is_bot")
        super().__init__(*args, **kwargs)
        self.ui.content.setReadOnly(True)
        self.ui.content.insertPlainText(self.message)
        self.ui.content.document().contentsChanged.connect(self.sizeChange)
        name = self.name
        if self.is_bot:
            self.ui.bot_name.show()
            self.ui.bot_name.setText(f"{name}")
            self.ui.bot_name.setStyleSheet("font-weight: normal;")
            self.ui.user_name.hide()
        else:
            self.ui.user_name.show()
            self.ui.user_name.setText(f"{name}")
            self.ui.user_name.setStyleSheet("font-weight: normal;")
            self.ui.bot_name.hide()

        self.ui.content.setStyleSheet("border-radius: 5px; border: 5px solid #1f1f1f; background-color: #1f1f1f; color: #ffffff;")

        self.register(SignalCode.APPLICATION_SETTINGS_CHANGED_SIGNAL, self.on_application_settings_changed_signal)
        self.font_family = None
        self.font_size = None
        self.set_chat_font()

    def on_application_settings_changed_signal(self):
        self.set_chat_font()

    def set_chat_font(self):
        font_setting = self.get_font_setting_by_name("chat")
        font_family = font_setting.font_family
        font_size = font_setting.font_size
        if self.font_family != font_family or self.font_size != font_size:
            self.font_family = font_family
            self.font_size = font_size
            # Check if the font family is available
            if self.font_family in QFontDatabase().families():
                font = QFont(self.font_family, self.font_size)
                font.setFamilies([
                    self.font_family,
                    "Noto Color Emoji",
                ])
                self.ui.content.setFont(font)

    def set_content_size(self):
        doc_height = self.ui.content.document().size().height()
        doc_width = self.ui.content.document().size().width()
        self.setMinimumHeight(int(doc_height) + 25)
        self.setMinimumWidth(int(doc_width))

    def sizeChange(self):
        self.set_content_size()
        self.textChanged.emit()

    def resizeEvent(self, event):
        self.set_content_size()
        super().resizeEvent(event)

    def sizeHint(self):
        fm = QFontMetrics(self.font())
        h = fm.height() * (self.ui.content.document().lineCount() + 1)
        return QSize(self.width(), h)

    def minimumSizeHint(self):
        return self.sizeHint()

    def update_message(self, text):
        self.message += text

        # strip double spaces from self.message
        self.message = self.message.replace("  ", " ")

        self.ui.content.setPlainText(self.message)




from PySide6.QtWidgets import QWizard
from airunner.mediator_mixin import MediatorMixin
from airunner.windows.installer.completion_page import CompletionPage
from airunner.windows.installer.confirmation_page import ConfirmationPage
from airunner.windows.installer.download_page import DownloadPage
from airunner.windows.main.settings_mixin import SettingsMixin


class InstallerWindow(
    QWizard,
    MediatorMixin,
    SettingsMixin
):
    def __init__(self, *args):
        MediatorMixin.__init__(self)
        SettingsMixin.__init__(self)
        super(InstallerWindow, self).__init__(*args)

        self.page_ids = {}

        self.do_download_sd_models = True
        self.do_download_controlnet_models = True
        self.do_download_llm = True
        self.do_download_tts_models = True
        self.do_download_stt_models = True

        self.pages = {
            "confirmation_page": ConfirmationPage(self),
            "download_page": DownloadPage(self),
            "completion_page": CompletionPage(self),
        }

        for page_name, page in self.pages.items():
            self.addPage(page)

        self.setWindowTitle("AI Runner Setup Wizard")

    @property
    def download_settings(self):
        return {
            "compile_with_pyinstaller": False,
            "download_ai_runner": True,
            "download_sd": self.do_download_sd_models,
            "download_controlnet": self.do_download_controlnet_models,
            "download_llm": self.do_download_llm,
            "download_tts": self.do_download_tts_models,
            "download_stt": self.do_download_stt_models,
        }

    def addPage(self, page):
        page_id = super().addPage(page)
        self.page_ids[page] = page_id
        return page_id